/*---------------
 * pg_decode_record
 *
 * FIXME
 *
 * This program is open source, licensed under the PostgreSQL license.
 * For license terms, see the LICENSE file.
 *
 * Copyright (C) 2023: Julien Rouhaud
 *
 */

#include "postgres.h"

#include "access/detoast.h"
#include "access/genam.h"
#if PG_VERSION_NUM < 120000
#include "access/heapam.h"
#endif
#include "access/heapam_xlog.h"
#include "access/htup.h"
#include "access/htup_details.h"
#include "access/stratnum.h"
#if PG_VERSION_NUM >= 120000
#include "access/table.h"
#endif
#include "catalog/catalog.h"
#if PG_VERSION_NUM < 140000
#include "catalog/indexing.h"
#endif
#include "catalog/pg_tablespace_d.h"
#include "fmgr.h"
#include "storage/item.h"
#include "utils/builtins.h"
#include "utils/fmgroids.h"
#include "utils/lsyscache.h"
#include "utils/rel.h"
#include "utils/relcache.h"
#include "utils/relmapper.h"
#if PG_VERSION_NUM >= 160000
#include "varatt.h"
#endif

PG_MODULE_MAGIC;

#if PG_VERSION_NUM < 120000
#define table_open(r, l)	heap_open(r, l)
#define table_close(r, l)	heap_close(r, l)
#endif

#if PG_VERSION_NUM < 160000
#define RelationMapFilenumberToOid(n, s) RelationMapFilenodeToOid(n, s)
#endif

/*--- Functions --- */

void	_PG_init(void);

extern PGDLLEXPORT Datum	pg_decode_record(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(pg_decode_record);

static void decode_heap_insert(bytea *record, Oid spcOid, Oid relNumber,
							   StringInfo buf);
static Oid get_spc_relnumber_relid(Oid spcOid, Oid relNumber);


void
_PG_init(void)
{
	/* nothing to do */
}

/*
 * For a given tablespace oid and relfilenode, find the underlying relation
 * oid.
 */
static Oid
get_spc_relnumber_relid(Oid spcOid, Oid relNumber)
{
	Relation	pg_class;
	ScanKeyData keys[2];
	SysScanDesc scan;
	HeapTuple	tuple;
	Oid			relid;

	ScanKeyInit(&keys[0],
				Anum_pg_class_reltablespace,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(spcOid));
	ScanKeyInit(&keys[1],
				Anum_pg_class_relfilenode,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(relNumber));

	pg_class = table_open(RelationRelationId, AccessShareLock);

	scan = systable_beginscan(pg_class, ClassTblspcRelfilenodeIndexId, true,
							  NULL, 2, keys);

	tuple = systable_getnext(scan);
	if (HeapTupleIsValid(tuple))
	{
#if PG_VERSION_NUM < 120000
		relid = HeapTupleGetOid(tuple);
#else
		relid = ((Form_pg_class) GETSTRUCT(tuple))->oid;
#endif
	}
	else
		relid = InvalidOid;
	systable_endscan(scan);

	table_close(pg_class, AccessShareLock);

	return relid;
}

static void
decode_heap_insert(bytea *record, Oid spcOid, Oid relNumber, StringInfo buf)
{
	Oid			relid;
	Relation	relation;
	TupleDesc	tupdesc;
	HeapTuple	tuple;
	Datum	   *values;
	bool	   *isnull;
	int			i;
	union
	{
		HeapTupleHeaderData hdr;
		char		data[MaxHeapTupleSize];
	}			tbuf;
	HeapTupleHeader htup;
	xl_heap_header xlhdr;
	uint32		newlen;
	Size		datalen;
	char	   *data;
	bool		first_col;

	relid = get_spc_relnumber_relid(spcOid, relNumber);

	/* Default tablespace has a 0 reltablespace */
	if (!OidIsValid(relid) && spcOid == DEFAULTTABLESPACE_OID)
		relid = get_spc_relnumber_relid(InvalidOid, relNumber);

	if (!OidIsValid(relid) && spcOid == GLOBALTABLESPACE_OID)
		relid = RelationMapFilenumberToOid(relNumber, true);

	if (!OidIsValid(relid))
		elog(ERROR, "could not find relation for tblspc %u relfilenode %u",
					spcOid, relNumber);

	relation = table_open(relid, AccessShareLock);
	tupdesc = CreateTupleDescCopy(RelationGetDescr(relation));

	initStringInfo(buf);
	appendStringInfo(buf, "INSERT INTO %s.%s",
					 quote_identifier(get_namespace_name(RelationGetNamespace(relation))),
					 quote_identifier(RelationGetRelationName(relation)));

	/* mimic heap_xlog_insert */
	data = VARDATA(record);
	datalen = VARSIZE_ANY(record);

	newlen = datalen - SizeOfHeapHeader;
	Assert(datalen > SizeOfHeapHeader && newlen <= MaxHeapTupleSize);
	memcpy((char *) &xlhdr, data, SizeOfHeapHeader);
	data += SizeOfHeapHeader;

	htup = &tbuf.hdr;
	MemSet((char *) htup, 0, SizeofHeapTupleHeader);
	/* PG73FORMAT: get bitmap [+ padding] [+ oid] + data */
	memcpy((char *) htup + SizeofHeapTupleHeader,
		   data,
		   newlen);
	newlen += SizeofHeapTupleHeader;
	htup->t_infomask2 = xlhdr.t_infomask2;
	htup->t_infomask = xlhdr.t_infomask;
	htup->t_hoff = xlhdr.t_hoff;

	/* build a fake tuple with the bare minimum to deform it */
	tuple = (HeapTuple) palloc0(HEAPTUPLESIZE + VARSIZE_ANY(record));
	tuple->t_data = htup;
	tuple->t_len = VARSIZE_ANY(record);
	ItemPointerSetInvalid(&(tuple->t_self));
	tuple->t_tableOid = relid;

	values = palloc0(sizeof(Datum) * tupdesc->natts);
	isnull = palloc0(sizeof(bool) * tupdesc->natts);
	heap_deform_tuple(tuple, tupdesc, values, isnull);

	/* append the column list, or the underlying table name for TOAST */
	if (!IsToastRelation(relation))
	{
		first_col = true;
		appendStringInfoString(buf, " (");
		for (i = 0; i < tupdesc->natts; i++)
		{
			if (TupleDescAttr(tupdesc, i)->attisdropped)
				continue;

			if (first_col)
				first_col = false;
			else
				appendStringInfoString(buf, ", ");

			appendStringInfo(buf, "%s",
							 quote_identifier(TupleDescAttr(tupdesc, i)->attname.data));
		}
		appendStringInfoString(buf, ")");
	}
	else
	{
		Oid		real_relid;

		Assert(strncmp(RelationGetRelationName(relation), "pg_toast_", 9) == 0);

		real_relid = atoi(RelationGetRelationName(relation) + 9);
		appendStringInfo(buf, " /* %s.%s */",
					 quote_identifier(get_namespace_name(get_rel_namespace(real_relid))),
					 quote_identifier(get_rel_name(real_relid)));
	}

	/* append the values */
	first_col = true;
	appendStringInfoString(buf, " VALUES (");
	for (i = 0; i < tupdesc->natts; i++)
	{
		char	   *value = NULL;
		Oid			typoutput;
		bool		typisvarlena;

		if (TupleDescAttr(tupdesc, i)->attisdropped)
			continue;

		if (first_col)
			first_col = false;
		else
			appendStringInfoString(buf, ", ");

		if (isnull[i])
		{
			appendStringInfoString(buf, "NULL");
			continue;
		}

		getTypeOutputInfo(TupleDescAttr(tupdesc, i)->atttypid,
						  &typoutput, &typisvarlena);

		if (typisvarlena)
		{
			Pointer p = DatumGetPointer(values[i]);

			if (VARATT_IS_EXTERNAL_ONDISK(p))
			{
				struct varlena *datum = (struct varlena *) p;
				struct varatt_external toast_pointer;

				VARATT_EXTERNAL_GET_POINTER(toast_pointer, datum);

				value = palloc(sizeof(char) * 255);

				snprintf(value, 255, "/* toast pointer %u */",
						 toast_pointer.va_valueid);
			}
		}

		if (!value)
		{
			value = OidOutputFunctionCall(typoutput, values[i]);
			value = quote_literal_cstr(value);
		}

		appendStringInfo(buf, "%s", value);

		pfree(value);
	}
	appendStringInfoString(buf, ");");

	table_close(relation, NoLock);
}

PGDLLEXPORT Datum
pg_decode_record(PG_FUNCTION_ARGS)
{
	bytea	   *record = PG_GETARG_BYTEA_PP(0);
	Oid			spc = PG_GETARG_OID(1);
	Oid			relfilenode = PG_GETARG_OID(2);
	StringInfoData buf;

	// quick hack: we know we only extracted HEAP INSERT records */
	decode_heap_insert(record, spc, relfilenode, &buf);

	PG_RETURN_TEXT_P(cstring_to_text(buf.data));
}
